/* (C) 2009-2010 Yana */

#undef TESTMIPS32
#undef TESTMIPS64
#include "device.h"

void device_cons_put_hexadecimal(unsigned int hexadecimal) {
	unsigned int i;
	for (i = 0; i < 8; ++i) {
		device_cons_putchar("0123456789abcdef"[hexadecimal >> (28 - i * 4) & 0xf]);
	}
}

void device_cons_put_string(const char *string) {
	while (*string != 0) {
		device_cons_putchar(*string);
		++string;
	}
}

#include "yk0a.h"
#include "yk0a_arm.h"
#include "yk0a_interrupt.h"
#include <stddef.h>

yk0a_uint32_t glue_arm_read_control(void) {
	yk0a_uint32_t control;
	__asm__ __volatile__("mrc p15, 0, %0, c1, c0, 0" : "=r"(control));
	return control;
}

void glue_arm_write_control(yk0a_uint32_t control) {
	__asm__ __volatile__("mcr p15, 0, %0, c1, c0, 0" : : "r"(control));
}

void glue_arm_write_translation(yk0a_uint32_t translation) {
	__asm__ __volatile__("mcr p15, 0, %0, c2, c0, 0" : : "r"(translation));
}

void glue_arm_write_domain(yk0a_uint32_t domain) {
	__asm__ __volatile__("mcr p15, 0, %0, c3, c0, 0" : : "r"(domain));
}

void glue_arm_tlb_invalidate(void) {
	__asm__ __volatile__("mcr p15, 0, %0, c8, c7, 0" : : "r"(0));
}

#include <signal.h>
#include <string.h>
#include "kernel.h"
#include "proc.h"

extern irq_hook_t *irq_handlers[NR_IRQ_VECTORS];

#define HANDLER_BIT 0x100
#define INTERRUPT_BIT (DEVICE_INTERRUPT_RTC | DEVICE_INTERRUPT_ETHER | DEVICE_INTERRUPT_CONS)
#define INTERRUPT_SIZE 8
#define LEVEL1_SIZE 4096 /* 4096 MB */
#define LEVEL1_UNIT 0x100000
#define LEVEL2_SIZE 2048 /* 8 MB */
#define REMOTE_SIZE 0x10000 /* 256 MB */
#define REMOTE_VIR 0x10000
#define STACK_BOOT 12
#define STACK_FRAME 17

#if !defined(MINIX_PROCESS) || MINIX_PROCESS % LEVEL1_UNIT != 0
#error MINIX_PROCESS
#endif

const yk0a_uint8_t glue_boot[] = "memory=00100000:00f00000\0rootdev=769";

volatile yk0a_uint32_t glue_interrupt_lock;
volatile yk0a_interrupt_t glue_interrupt_mask;
volatile yk0a_interrupt_t glue_interrupt_request;

yk0a_uint32_t glue_level1[LEVEL1_SIZE] __attribute__((aligned(16384)));
yk0a_uint32_t glue_level2[LEVEL2_SIZE] __attribute__((aligned(1024)));

void glue_exception(int signal) {
	if (iskernelp(proc_ptr)) {
		panic("task exception", signal);
	} else {
		cause_sig(proc_nr(proc_ptr), signal);
	}
}

void glue_task(phys_bytes a) {
	yk0a_uint32_t i;
	for (i = 0; i < LEVEL1_SIZE; ++i) {
		if (i * (LEVEL1_UNIT / CLICK_SIZE) >= REMOTE_VIR && i * (LEVEL1_UNIT / CLICK_SIZE) < REMOTE_VIR + REMOTE_SIZE) {
			glue_level1[i] = i << 20 | 0x422; /* AP = 1, domain = 1, C = 0, type = 2 */
		} else if (i * (LEVEL1_UNIT / CLICK_SIZE) >= MINIX_PROCESS / CLICK_SIZE && i * (LEVEL1_UNIT / CLICK_SIZE) < MINIX_PROCESS / CLICK_SIZE + LEVEL2_SIZE) {
			glue_level1[i] = (yk0a_uint32_t)&glue_level2[i * (LEVEL1_UNIT / CLICK_SIZE) - MINIX_PROCESS / CLICK_SIZE] | 1; /* domain = 0, type = 1 */
		} else {
			glue_level1[i] = i << 20 | 0x40a; /* AP = 1, domain = 0, C = 1, type = 2 */
		}
	}
	glue_arm_write_translation((yk0a_uint32_t)glue_level1);
	glue_arm_write_domain(0x5); /* D1 = 1, D0 = 1 */
	glue_arm_write_control(glue_arm_read_control() | 1); /* M = 1 */
	aout = a;
	cstart(0, 0, (phys_bytes)glue_boot >> 16, (phys_bytes)glue_boot, sizeof(glue_boot));
	main();
	yk0a_set_context(2, YK0A_CONTEXT_CPSR, 0, NULL);
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		if (next_ptr != NULL) {
			proc_ptr = next_ptr;
			next_ptr = NULL;
			glue_arm_tlb_invalidate();
			for (i = 0; i < LEVEL2_SIZE; ++i) {
				if (i + MINIX_PROCESS / CLICK_SIZE >= proc_ptr->p_memmap[T].mem_vir && i + MINIX_PROCESS / CLICK_SIZE < proc_ptr->p_memmap[T].mem_vir + proc_ptr->p_memmap[T].mem_len) {
					glue_level2[i] = (i + MINIX_PROCESS / CLICK_SIZE + proc_ptr->p_memmap[T].mem_phys - proc_ptr->p_memmap[T].mem_vir) * CLICK_SIZE | 0xaaa; /* AP3 = 2, AP2 = 2, AP1 = 2, AP0 = 2, C = 1, type = 2 */
				} else if (i + MINIX_PROCESS / CLICK_SIZE >= proc_ptr->p_memmap[D].mem_vir && i + MINIX_PROCESS / CLICK_SIZE < proc_ptr->p_memmap[S].mem_vir + proc_ptr->p_memmap[S].mem_len) {
					glue_level2[i] = (i + MINIX_PROCESS / CLICK_SIZE + proc_ptr->p_memmap[D].mem_phys - proc_ptr->p_memmap[D].mem_vir) * CLICK_SIZE | 0xffa; /* AP3 = 3, AP2 = 3, AP1 = 3, AP0 = 3, C = 1, type = 2 */
				} else {
					glue_level2[i] = 0; /* type = 0 */
				}
			}
			if (proc_ptr->p_priv->s_farmem[0].in_use != 0) {
				glue_arm_write_domain(0xd); /* D1 = 3, D0 = 1 */
			} else {
				glue_arm_write_domain(0x5); /* D1 = 1, D0 = 1 */
			}
#if defined(YK0A_CUSTOM_ENABLE_PORT)
			if (iskernelp(proc_ptr)) {
				yk0a_set_context(2, YK0A_CUSTOM_CONTEXT_PORT, 0, NULL);
			} else {
				yk0a_set_context(2, YK0A_CUSTOM_CONTEXT_PORT, 1, NULL);
			}
#endif
		}
		for (i = 0; i < STACK_FRAME; ++i) {
			yk0a_set_context(2, i, ((yk0a_uint32_t *)&proc_ptr->p_reg)[i], NULL);
		}
		yk0a_reply(2);
		yk0a_listen(1, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) == 0) {
			yk0a_forward(1, 2);
		}
		if (yk0a_accept(1, &t) != 0 || t != 2) {
device_cons_put_string("{accept}\n");
while (1) {
}
		}
		for (i = 0; i < STACK_FRAME; ++i) {
			yk0a_get_context(2, i, &((yk0a_uint32_t *)&proc_ptr->p_reg)[i]);
		}
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0) {
			yk0a_uint32_t cpsr;
			yk0a_set_context(2, YK0A_CONTEXT_CPSR, 0, &cpsr);
			switch ((cpsr & YK0A_ARM_CPSR_M_MASK) >> YK0A_ARM_CPSR_M_SHIFT) {
			case 0:
#if defined(YK0A_CUSTOM_ENABLE_PORT)
				if (iskernelp(proc_ptr)) {
					proc_ptr->p_reg.retreg = sys_call(proc_ptr->p_reg.r2, proc_ptr->p_reg.r3, (message *)proc_ptr->p_reg.r4, 0);
				} else {
					proc_ptr->p_reg.retreg = sys_call(proc_ptr->p_reg.retreg, proc_ptr->p_reg.r1, (message *)proc_ptr->p_reg.r2, 0);
				}
#else
				proc_ptr->p_reg.retreg = sys_call(proc_ptr->p_reg.r2, proc_ptr->p_reg.r3, (message *)proc_ptr->p_reg.r4, 0);
#endif
				break;
			case 0x17: /* abt */
				glue_exception(SIGSEGV);
				break;
			case 0x1b: /* und */
				glue_exception(SIGILL);
				break;
			default:
device_cons_put_string("{exception ");
device_cons_put_hexadecimal(cpsr);
device_cons_put_string("}\n");
while (1) {
}
				break;
			}
		}
		if ((f & ~YK0A_FLAG_KERNEL_MASK) != 0) {
			glue_interrupt_request |= f & glue_interrupt_mask;
			if (glue_interrupt_lock == 0 && ((glue_interrupt_mask & glue_interrupt_request) != 0)) {
				for (i = INTERRUPT_SIZE - 1; i > 0; --i) {
					if ((1 << i & glue_interrupt_mask & glue_interrupt_request) != 0) {
						glue_interrupt_mask &= ~(1 << i);
						glue_interrupt_request &= ~(1 << i);
						intr_handle(irq_handlers[i]);
						if (1 << i == DEVICE_INTERRUPT_RTC) {
							*(volatile yk0a_uint32_t *)DEVICE_RTC_INTERRUPT_ACK;
						}
						if (irq_actids[i] == 0) {
							glue_interrupt_mask |= 1 << i;
							yk0a_interrupt_bit_mask(0, 1 << i, NULL);
						}
					}
				}
			}
		}
	}
}

unsigned inb(U16_t _port) {
device_cons_put_string("{inb}\n");
while (1) {
}
}

unsigned inw(U16_t _port) {
device_cons_put_string("{inw}\n");
while (1) {
}
}

unsigned inl(U32_t _port) {
device_cons_put_string("{inl}\n");
while (1) {
}
}

void outb(U16_t _port, U8_t _value) {
device_cons_put_string("{outb}\n");
while (1) {
}
}

void outw(U16_t _port, U16_t _value) {
device_cons_put_string("{outw}\n");
while (1) {
}
}

void outl(U16_t _port, U32_t _value) {
device_cons_put_string("{outl}\n");
while (1) {
}
}

void intr_disable(void) {
	yk0a_interrupt_t i;
	yk0a_interrupt_bit_mask(INTERRUPT_BIT, 0, &i);
	++glue_interrupt_lock;
	yk0a_interrupt_bit_mask(0, i & INTERRUPT_BIT, NULL);
}

void intr_enable(void) {
	yk0a_uint32_t b;
	yk0a_interrupt_t i;
	b = 0;
	yk0a_interrupt_bit_mask(INTERRUPT_BIT, 0, &i);
	if (glue_interrupt_lock != 0) {
		--glue_interrupt_lock;
		if (glue_interrupt_lock == 0 && (glue_interrupt_mask & glue_interrupt_request) != 0) {
			b = 1;
		}
	}
	yk0a_interrupt_bit_mask(0, i & INTERRUPT_BIT, NULL);
	if (b != 0) {
		yk0a_bit_flag(1, 0, HANDLER_BIT, NULL);
	}
}

void phys_copy(phys_bytes source, phys_bytes dest, phys_bytes count) {
	memcpy((void *)dest, (void *)source, count);
}

void alloc_segments(struct proc *rp) {
	/* ToDo: MMU */
	if (iskernelp(rp)) {
		rp->p_memmap[T].mem_vir = rp->p_memmap[T].mem_phys;
		rp->p_memmap[D].mem_vir = rp->p_memmap[D].mem_phys;
		rp->p_memmap[S].mem_vir = rp->p_memmap[S].mem_phys;
		rp->p_reg.sp -= STACK_BOOT;
	} else {
		rp->p_memmap[T].mem_vir = MINIX_PROCESS / CLICK_SIZE;
		rp->p_memmap[D].mem_vir = rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len;
		rp->p_memmap[S].mem_vir = rp->p_memmap[D].mem_vir + rp->p_memmap[S].mem_phys - rp->p_memmap[D].mem_phys;
		if (rp->p_reg.sp == 0) {
			rp->p_reg.sp = (rp->p_memmap[S].mem_vir + rp->p_memmap[S].mem_len) * CLICK_SIZE - STACK_BOOT;
		}
	}
}

void vm_init(void) {
device_cons_put_string("{vm_init}\n");
while (1) {
}
}

void vm_map_range(u32_t base, u32_t size, u32_t offset) {
device_cons_put_string("{vm_map_range}\n");
while (1) {
}
}

phys_bytes umap_local(register struct proc *rp, int seg, vir_bytes vir_addr, vir_bytes bytes) {
	if (seg == T) {
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[T].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len) {
			return vir_addr + (rp->p_memmap[T].mem_phys - rp->p_memmap[T].mem_vir) * CLICK_SIZE;
		}
	} else {
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[T].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len) { /* ToDo: .rodata */
			return vir_addr + (rp->p_memmap[T].mem_phys - rp->p_memmap[T].mem_vir) * CLICK_SIZE;
		}
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[D].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[S].mem_vir + rp->p_memmap[S].mem_len) {
			return vir_addr + (rp->p_memmap[D].mem_phys - rp->p_memmap[D].mem_vir) * CLICK_SIZE;
		}
	}
	return 0;
}

void cp_mess(int src, phys_clicks src_clicks, vir_bytes src_offset, phys_clicks dst_clicks, vir_bytes dst_offset) {
	message *m;
	m = (message *)(dst_clicks * CLICK_SIZE + dst_offset);
	memcpy(m, (void *)(src_clicks * CLICK_SIZE + src_offset), sizeof(message));
	m->m_source = src;
}

phys_bytes umap_remote(struct proc* rp, int seg, vir_bytes vir_addr, vir_bytes bytes) {
device_cons_put_string("{umap_remote}\n");
while (1) {
}
}

phys_bytes seg2phys(U16_t seg) {
device_cons_put_string("{seg2phys}\n");
while (1) {
}
}

void phys_memset(phys_bytes source, unsigned long pattern, phys_bytes count) {
	memset((void *)source, pattern, count);
}

vir_bytes alloc_remote_segment(u32_t *selector, segframe_t *segments, int index, phys_bytes phys, vir_bytes size, int priv) {
	/* ToDo: remote segment */
	return phys;
}

int arch_init_clock(void) {
	*(volatile yk0a_uint32_t *)DEVICE_RTC_HZ = HZ;
	return 0;
}

#if 0
clock_t read_clock(void) {
device_cons_put_string("{read_clock}\n");
while (1) {
}
}
#else /* patch */
#endif

void clock_stop(void) {
device_cons_put_string("{clock_stop}\n");
while (1) {
}
}

int intr_init(int mine) {
	return 0;
}

int intr_disabled(void) {
	return glue_interrupt_lock;
}

int intr_unmask(irq_hook_t* hook) {
	irq_actids[hook->irq] &= ~hook->id;
	if (irq_actids[hook->irq] == 0) {
		yk0a_uint32_t b;
		yk0a_interrupt_t i;
		b = 0;
		yk0a_interrupt_bit_mask(INTERRUPT_BIT, 0, &i);
		i |= 1 << hook->irq;
		glue_interrupt_mask |= 1 << hook->irq;
		if (glue_interrupt_lock == 0 && (glue_interrupt_mask & glue_interrupt_request) != 0) {
			b = 1;
		}
		yk0a_interrupt_bit_mask(0, i & INTERRUPT_BIT, NULL);
		if (b != 0) {
			yk0a_bit_flag(1, 0, HANDLER_BIT, NULL);
		}
	}
	return 0;
}

int intr_mask(irq_hook_t* hook) {
	yk0a_interrupt_t i;
	irq_actids[hook->irq] |= hook->id;
	yk0a_interrupt_bit_mask(INTERRUPT_BIT, 0, &i);
	i &= ~(1 << hook->irq);
	glue_interrupt_mask &= ~(1 << hook->irq);
	yk0a_interrupt_bit_mask(0, i & INTERRUPT_BIT, NULL);
	return 0;
}

void idle_task(void) {
	while (1) {
	}
}

void system_init(void) {
}

void ser_putc(char c) {
device_cons_put_string("{ser_putc}\n");
while (1) {
}
}

void arch_shutdown(int how) {
device_cons_put_string("{arch_shutdown}\n");
while (1) {
}
}

void restart(void) {
device_cons_put_string("{restart}\n");
while (1) {
}
}

void read_tsc(unsigned long *high, unsigned long *low) {
	/* ToDo: TSC */
}

int arch_init_profile_clock(u32_t freq) {
device_cons_put_string("{arch_init_profile_clock}\n");
while (1) {
}
}

void arch_stop_profile_clock(void) {
device_cons_put_string("{arch_stop_profile_clock}\n");
while (1) {
}
}

void arch_ack_profile_clock(void) {
device_cons_put_string("{arch_ack_profile_clock}\n");
while (1) {
}
}

int do_int86(message *m_ptr) {
device_cons_put_string("{do_int86}\n");
while (1) {
}
}

int do_sdevio(message *m_ptr) {
device_cons_put_string("{do_sdevio}\n");
while (1) {
}
}

int do_iopenable(message *m_ptr) {
device_cons_put_string("{do_iopenable}\n");
while (1) {
}
}

int do_readbios(message *m_ptr) {
device_cons_put_string("{do_readbios}\n");
while (1) {
}
}
