/* (C) 2009-2010 Yana */

#include "device.h"
#include "start_setup.h"
#include "tool.h"
#include "yk0a.h"
#include "yk0a_setup.h"
#include <stddef.h>

yk0a_uint8_t task0_stack[TASK0_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));

yk0a_uint32_t disk_read(yk0a_uint32_t id, yk0a_uint32_t offset, void *address, yk0a_uint32_t size) {
	yk0a_uint32_t i;
	for (i = 0; i < size; ++i) {
		yk0a_uint32_t a;
		if (i % DEVICE_DISK_BUFFER_LEN == 0) {
			if (size - i >= DEVICE_DISK_BUFFER_LEN) {
				a = 0;
			} else {
				if (DEVICE_DISK_BUFFER_LEN - (size - i) > i + offset) {
					a = i + offset;
				} else {
					a = DEVICE_DISK_BUFFER_LEN - (size - i);
				}
			}
			*(volatile yk0a_uint32_t *)DEVICE_DISK_OFFSET = i + offset - a;
			*(volatile yk0a_uint32_t *)DEVICE_DISK_OFFSET_HIGH32 = 0;
			*(volatile yk0a_uint32_t *)DEVICE_DISK_ID = id;
			*(volatile yk0a_uint32_t *)DEVICE_DISK_START_OPERATION = DEVICE_DISK_OPERATION_READ;
			if (*(volatile yk0a_uint32_t *)DEVICE_DISK_STATUS == 0) {
				return i;
			}
		}
		((yk0a_uint8_t *)address)[i] = ((volatile yk0a_uint8_t *)DEVICE_DISK_BUFFER)[i % DEVICE_DISK_BUFFER_LEN + a];
	}
	return size;
}

#include <minix/com.h>
#include <minix/config.h>
#include <minix/const.h>

#include <a.out.h>

#define NAME_SIZE 64

#if !defined(MINIX_KERNEL) || MINIX_KERNEL % CLICK_SIZE != 0
#error MINIX_KERNEL
#endif

struct exec aout[INIT_PROC_NR + 2];

void task0(void) {
	yk0a_uint32_t a;
	yk0a_uint32_t i;
	yk0a_uint32_t o;
	a = MINIX_KERNEL;
	o = 0;
	for (i = 0; i < INIT_PROC_NR + 2; ++i) {
		o += NAME_SIZE;
		if (disk_read(0, o, &aout[i], sizeof(struct exec)) != sizeof(struct exec)) {
			tool_cons_put_string("{");
			tool_cons_put_hexadecimal(o);
			tool_cons_put_string(" ");
			tool_cons_put_hexadecimal(sizeof(struct exec));
			tool_cons_put_string("}\n");
			device_cons_halt();
		}
		o += sizeof(struct exec);
		aout[i].a_syms = a;
		if (disk_read(0, o, (void *)a, aout[i].a_text) != aout[i].a_text) {
			tool_cons_put_string("{");
			tool_cons_put_hexadecimal(o);
			tool_cons_put_string(" ");
			tool_cons_put_hexadecimal(aout[i].a_text);
			tool_cons_put_string("}\n");
			device_cons_halt();
		}
		a += aout[i].a_text;
		o += aout[i].a_text;
		if (disk_read(0, o, (void *)a, aout[i].a_data) != aout[i].a_data) {
			tool_cons_put_string("{");
			tool_cons_put_hexadecimal(o);
			tool_cons_put_string(" ");
			tool_cons_put_hexadecimal(aout[i].a_data);
			tool_cons_put_string("}\n");
			device_cons_halt();
		}
		a = (a + aout[i].a_total + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE;
		o += aout[i].a_data;
	}
	yk0a_set_priority(1, 1, NULL);
	for (i = 2; i < YK0A_SETUP_TASK_SIZE; ++i) {
		yk0a_connect(i, 1);
		yk0a_set_priority(i, 2, NULL);
	}
	yk0a_set_context(1, YK0A_CONTEXT_R0, (yk0a_uint32_t)aout, NULL);
	tool_task_start(1, (void (*)(void))aout[0].a_entry, NULL);
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		yk0a_listen(YK0A_SETUP_EXCEPTION_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_EXCEPTION_PORT, &t) == 0) {
			if (t != 1) {
				yk0a_forward(1, t);
			} else {
				yk0a_uint32_t cpsr;
				yk0a_get_context(t, YK0A_CONTEXT_CPSR, &cpsr);
				tool_cons_put_string("{");
				tool_cons_put_hexadecimal(f);
				tool_cons_put_string(" ");
				tool_cons_put_hexadecimal(t);
				tool_cons_put_string(" ");
				tool_cons_put_hexadecimal(cpsr);
				tool_cons_put_string("}\n");
				device_cons_halt();
			}
		}
		if ((f & ~YK0A_FLAG_KERNEL_MASK) != 0) {
			yk0a_bit_flag(1, 0, f & ~YK0A_FLAG_KERNEL_MASK, NULL);
		}
	}
}
